﻿-- Author      : RisM
-- Create Date : 6/28/2009 3:57:34 PM

SpeakinSpell = LibStub("AceAddon-3.0"):GetAddon("SpeakinSpell")
local L					= LibStub("AceLocale-3.0"):GetLocale("SpeakinSpell",					false)
local HELPFILE			= LibStub("AceLocale-3.0"):GetLocale("SpeakinSpell_HELPFILE",			false)
local DEFAULT_SPEECHES	= LibStub("AceLocale-3.0"):GetLocale("SpeakinSpell_DEFAULT_SPEECHES",	false)
local BIGREDBUTTON		= LibStub("AceLocale-3.0"):GetLocale("SpeakinSpell_BIGREDBUTTON",		false)

-------------------------------------------------------------------------------
-- OPTIONS GUI FUNCTIONS - GENERAL SETTINGS
-------------------------------------------------------------------------------


function SpeakinSpell:GeneralOptions_OnShowSpellEventsToggle(getset, value)
	if "GET" == getset then
		return SpeakinSpellSavedData.ShowSpellEvents
	else -- "SET"
		SpeakinSpellSavedData.ShowSpellEvents = value
	end
end


function SpeakinSpell:GeneralOptions_OnShowDebugMessagesToggle(getset, value)
	if "GET" == getset then
		return SpeakinSpellSavedData.ShowDebugMessages
	else -- "SET"
		SpeakinSpellSavedData.ShowDebugMessages = value
	end
end


function SpeakinSpell:GeneralOptions_OnEnableAllMessagesToggle(getset, value)
	if "GET" == getset then
		return SpeakinSpellSavedData.EnableAllMessages
	else -- "SET"
		SpeakinSpellSavedData.EnableAllMessages = value
	end
end


function SpeakinSpell:GeneralOptions_OnClickImport()
	self:ImportNewDefaults()
end


function SpeakinSpell:GeneralOptions_OnClickBigRedButton()
	-- check range of BigRedButtonIndex
	if not self.RuntimeData.BigRedButtonIndex then
		self.RuntimeData.BigRedButtonIndex = 1
	end
	if self.RuntimeData.BigRedButtonIndex > #(BIGREDBUTTON.List) then
		self.RuntimeData.BigRedButtonIndex = 1
	end
	
	-- process substitution variables
	local DetectedEventStub = {
		name = "Big Red Button",
		rank = "Silly",
		target = self:GetDefaultTarget(false),
		caster = UnitName("player"),
		type = "MACRO",
		linenumber = tostring(self.RuntimeData.BigRedButtonIndex),
	}
	local DetectedEvent = self:CreateDetectedEvent( DetectedEventStub )
	local msg = self:ProcessSubstitutionVars( BIGREDBUTTON.List[ self.RuntimeData.BigRedButtonIndex ], DetectedEvent )
	
	-- print the next message in the looping array
	if msg ~= "" then
		self:SayOneLine( msg, "SPEAKINSPELL CHANNEL", nil, DetectedEvent )
	end
	
	-- prepare for the next pass
	self.RuntimeData.BigRedButtonIndex = self.RuntimeData.BigRedButtonIndex + 1
end


-------------------------------------------------------------------------------
-- OPTIONS GUI FUNCTIONS - MESSAGE SETTINGS
-------------------------------------------------------------------------------


function SpeakinSpell:CurrentMessagesGUI_GetSelectedEventObject()
	local eventkey = self.OptionsGUIStates.MessageSettings.SelectedEventKey
	if eventkey then
		return SpeakinSpellSavedData.EventTable[eventkey]
	else
		return nil
	end
end


function SpeakinSpell:CurrentMessagesGUI_ValidateSelectedEvent()
	-- get the selected EventTableEntry
	local EventTableEntry = self:CurrentMessagesGUI_GetSelectedEventObject()
	if EventTableEntry then
		if self:MatchesFilter(EventTableEntry.DetectedEvent,true) then
			-- the current selection is valid, so keep it
			return
		end
		--else, the current selection did not match the filter, so select a default
	end
	
	-- select the first thing in the list, if possible
	for key,EventTableEntry in pairs(SpeakinSpellSavedData.EventTable) do
		if self:MatchesFilter(EventTableEntry.DetectedEvent,true) then
			-- the current selection is valid, so keep it
			self.OptionsGUIStates.MessageSettings.SelectedEventKey = key
			return
		end
	end
	
	-- our spell list is empty or nothing matches the filter
	self.OptionsGUIStates.MessageSettings.SelectedEventKey = nil
end



function SpeakinSpell:CurrentMessagesGUI_OnClickDelete()
	local funcname = "CurrentMessagesGUI_OnClickDelete"

	-- get the selected EventTableEntry
	local eventkey = self.OptionsGUIStates.MessageSettings.SelectedEventKey
	if not eventkey then
		self:ErrorMsg(funcname, "no event selected")
		return
	end
	
	-- delete the current selection from the SavedData table
	self:DeleteSpell(eventkey)
end


function SpeakinSpell:CurrentMessagesGUI_OnClickGotoCreate()
	InterfaceOptionsFrame_OpenToCategory(L["Create New..."])
end


function SpeakinSpell:CurrentMessagesGUI_OnEventTypeFilterSelect(getset, value)
	if "GET" == getset then
		-- rebuild the list to match the new filter if its out of date
		self:CurrentMessagesGUI_RebuildSpellList()
		return SpeakinSpell.RuntimeData.SelectedEventTypeFilter
	else -- "SET"
		-- if the filter is changing, then rebuild the spell/event list
		self:SetFilter(value,nil)
		-- rebuild the list to match the new filter
		self:CurrentMessagesGUI_RebuildSpellList()
	end
	-- NOTE: the current selection will be repaired to match the new filter in _ValidateSelectedEvent
	--		as a side effect when the GUI framework automatically tries to get the new value of 
	--		the rest of the controls on the page, including the selected event
end


function SpeakinSpell:CurrentMessagesGUI_OnGetSetEventTextFilter(getset, value)
	if "GET" == getset then
		-- rebuild the list to match the new filter if its out of date
		self:CurrentMessagesGUI_RebuildSpellList()
		return SpeakinSpell.RuntimeData.SelectedEventTextFilter
	else -- "SET"
		-- if the filter is changing, then rebuild the spell/event list
		self:SetFilter(nil,value)
		-- rebuild the list to match the new filter
		self:CurrentMessagesGUI_RebuildSpellList()
	end
	-- NOTE: the current selection will be repaired to match the new filter in _ValidateSelectedEvent
	--		as a side effect when the GUI framework automatically tries to get the new value of 
	--		the rest of the controls on the page, including the selected event
end

function SpeakinSpell:CurrentMessagesGUI_OnEventSelect(getset, value)
	if "GET" == getset then
		-- create a side-effect to update the shown/hidden state of all options GUI controls
		self:CurrentMessagesGUI_RebuildSpellList()
		-- return the current spell selection
		return self.OptionsGUIStates.MessageSettings.SelectedEventKey
	else -- "SET"
		-- update the current selection
		self.OptionsGUIStates.MessageSettings.SelectedEventKey = value
	end
end


function SpeakinSpell:CurrentMessagesGUI_OnChannelSelect(getset, scenario, value)
	local funcname = "CurrentMessagesGUI_OnChannelSelect"
		
	-- get the selected EventTableEntry
	local EventTableEntry = self:CurrentMessagesGUI_GetSelectedEventObject()
	if not EventTableEntry then
		self:ErrorMsg(funcname, "no event selected")
		return
	end

	if "GET" == getset then
		--self:DebugMsg(funcname,"GET "..scenario..":"..tostring(EventTableEntry.Channels[scenario]))
		if EventTableEntry.Channels[scenario] then
			return L[EventTableEntry.Channels[scenario]]
		else
			return L["Silent"]
		end
	else --"SET"
		-- store the new channel selection
		if value == L["Silent"] then
			EventTableEntry.Channels[scenario] = nil
		else
			-- create a reverse lookup table to un-localize channel names
			-- so we can return a channel name that is compatible with SendChatMessage()
			local CHANNEL_TABLE = {
				[L["SAY"]]			= "SAY",
				[L["PARTY"]]		= "PARTY",
				[L["RAID"]]			= "RAID",
				[L["EMOTE"]]		= "EMOTE",
				[L["BATTLEGROUND"]]	= "BATTLEGROUND",
				[L["YELL"]]			= "YELL",
				[L["RAID_WARNING"]]	= "RAID_WARNING",
				[L["GUILD"]]		= "GUILD",
				[L["SPEAKINSPELL CHANNEL"]] = "SPEAKINSPELL CHANNEL",
			}
			EventTableEntry.Channels[scenario] = CHANNEL_TABLE[value]
		end
		--self:DebugMsg(funcname,"set "..scenario..":"..tostring(EventTableEntry.Channels[scenario]))
	end
end



function SpeakinSpell:CurrentMessagesGUI_OnWhisperTarget(getset, value)
	-- make sure we have a valid spell selection
	local EventTableEntry = self:CurrentMessagesGUI_GetSelectedEventObject()
	if not EventTableEntry then
		self:ErrorMsg( "CurrentMessagesGUI_OnWhisperTarget", "no event selected" )
		return false
	end
	
	if "GET" == getset then
		return EventTableEntry.WhisperTarget
	else
		EventTableEntry.WhisperTarget = value
	end
end


function SpeakinSpell:CurrentMessagesGUI_OnMessage(getset, index, value)
	local EventTableEntry = self:CurrentMessagesGUI_GetSelectedEventObject()
	if not EventTableEntry then
		self:ErrorMsg("CurrentMessagesGUI_OnMessage", "no event selected")
		return ""
	end
	
	if "GET" == getset then
		local msg = EventTableEntry.Messages[index]
		if not msg then
			msg = ""
		end
		return msg
	else -- "SET"
		-- make sure we have a message table for the currently selected spell
		if not EventTableEntry.Messages then
			EventTableEntry.Messages = {}
		end
		
		-- store the new value into the message table
		EventTableEntry.Messages[index] = value
		self:DebugMsg( nil,"Message " .. tostring(index) .. " for spell " .. EventTableEntry.DetectedEvent.DisplayName .. " set to: " .. tostring(EventTableEntry.Messages[index]) )
		
		-- remove holes from the message table
		-- if the user has just deleted one of the entries, move everything else up to fill in the gap
		if value == "" then
			local Messages = EventTableEntry.Messages -- create reference for shortcut notation
			for i=index+1,SpeakinSpell.MAX.MESSAGES_PER_SPELL,1 do
				Messages[i-1] = Messages[i]
			end
			Messages[SpeakinSpell.MAX.MESSAGES_PER_SPELL] = nil
			-- destroy empty strings from the table
			for i=SpeakinSpell.MAX.MESSAGES_PER_SPELL-1,1,-1 do
				if Messages[i] == "" then
					Messages[i] = nil
				else
					break --the rest should be valid non-empty strings
				end
			end
		end
	end
end




function SpeakinSpell:CurrentMessagesGUI_OnSpellDisable(getset, value)
	local EventTableEntry = self:CurrentMessagesGUI_GetSelectedEventObject()
	if not EventTableEntry then
		self:ErrorMsg("CurrentMessagesGUI_OnSpellDisable", "no event selected")
		return
	end

	if "GET" == getset then
		return EventTableEntry.DisableAnnouncements
	else -- "SET"
		EventTableEntry.DisableAnnouncements = value
	end
end


function SpeakinSpell:CurrentMessagesGUI_OnFrequencySelect(getset, value)
	local EventTableEntry = self:CurrentMessagesGUI_GetSelectedEventObject()
	if not EventTableEntry then
		self:ErrorMsg("CurrentMessagesGUI_OnFrequencySelect", "no event selected")
		return
	end
	
	if "GET" == getset then	
		return EventTableEntry.Frequency
	else --"SET"
		EventTableEntry.Frequency = value
	end
end


function SpeakinSpell:CurrentMessagesGUI_OnCooldownSlider(getset, value)
	local funcname = "CurrentMessagesGUI_OnCooldownSlider"

	local EventTableEntry = self:CurrentMessagesGUI_GetSelectedEventObject()
	if not EventTableEntry then
		self:ErrorMsg(funcname, "no event selected")
		return
	end
	
	if "GET" == getset then	
		return EventTableEntry.Cooldown
	else --"SET"
		EventTableEntry.Cooldown = value
	end
end


function SpeakinSpell:CurrentMessagesGUI_OnOncePerCombatToggle(getset, value)
	local funcname = "CurrentMessagesGUI_OnOncePerCombatToggle"

	local EventTableEntry = self:CurrentMessagesGUI_GetSelectedEventObject()
	if not EventTableEntry then
		self:ErrorMsg(funcname, "no event selected")
		return
	end
	
	if "GET" == getset then	
		return EventTableEntry.OncePerCombat
	else --"SET"
		EventTableEntry.OncePerCombat = value
	end
end


function SpeakinSpell:CurrentMessagesGUI_OnOncePerTargetToggle(getset, value)
	local funcname = "CurrentMessagesGUI_OnOncePerTargetToggle"

	local EventTableEntry = self:CurrentMessagesGUI_GetSelectedEventObject()
	if not EventTableEntry then
		self:ErrorMsg(funcname, "no event selected")
		return
	end
	
	if "GET" == getset then	
		return EventTableEntry.OncePerTarget
	else --"SET"
		EventTableEntry.OncePerTarget = value
	end
end


function SpeakinSpell:CurrentMessagesGUI_ShowSpeechBox(i)	
	-- determine how many Spell Message N input boxes are shown
	-- typically we'll show at least 1 speech message input box if we have a spell selected
	-- we want to show 1 more than the number of speeches currently configured
	-- unless we don't have a spell selected or we're in EnterNewSpell mode
	local NumSpellMessageInputsShown = 0
	local EventTableEntry = self:CurrentMessagesGUI_GetSelectedEventObject()
	if EventTableEntry then
		NumSpellMessageInputsShown = 1 + #(EventTableEntry.Messages)
	end

	return (i <= NumSpellMessageInputsShown)
end


function SpeakinSpell:CurrentMessagesGUI_ShowSelectedEventControls()
	local EventTableEntry = self:CurrentMessagesGUI_GetSelectedEventObject()
	return (EventTableEntry ~= nil)
end


function SpeakinSpell:CurrentMessagesGUI_RebuildSpellList()
	if not self.OptionsGUIStates then
		return
	end
	if not self.OptionsGUIStates.MessageSettings.FilterChanged then
		return
	end
	
	-- reset the spell list
	self.OptionsGUI.args.CurrentMessagesGUI.args.EventSelect.values = {}
	local OldSelectedEventTableEntry = SpeakinSpellSavedData.EventTable[ self.OptionsGUIStates.MessageSettings.SelectedEventKey ]
	
	-- search all entries in the event table for events that match the current filter
	local anykey = nil
	for key,EventTableEntry in pairs(SpeakinSpellSavedData.EventTable) do
		if self:MatchesFilter(EventTableEntry.DetectedEvent,true) then
			-- add to the list of shown values
			self.OptionsGUI.args.CurrentMessagesGUI.args.EventSelect.values[key] = EventTableEntry.DetectedEvent.DisplayName
			-- make sure we have a current spell selected for the options GUI 
			-- we'll use the last one we see because it's easier
			anykey = key
		end
	end
	
	if OldSelectedEventTableEntry and self:MatchesFilter( OldSelectedEventTableEntry.DetectedEvent,true ) then
		self.OptionsGUIStates.MessageSettings.SelectedEventKey = OldSelectedEventTableEntry.DetectedEvent.key
	else
		self.OptionsGUIStates.MessageSettings.SelectedEventKey = anykey
	end
	
	self.OptionsGUIStates.MessageSettings.FilterChanged = false
end


-------------------------------------------------------------------------------
-- OPTIONS GUI FUNCTIONS - CREATE NEW
-------------------------------------------------------------------------------


function SpeakinSpell:CreateNew_GetSelectedEventObject()
	local eventkey = self.OptionsGUIStates.CreateNew.SelectedEventKey
	if eventkey then
		return SpeakinSpellSavedData.NewEventsDetected[eventkey]
	else
		return nil
	end
end


function SpeakinSpell:CreateNew_ValidateSelectedEvent()
	local DetectedEvent = self:CreateNew_GetSelectedEventObject()

	-- rebuild the list to match the new filter if its out of date
	self:CreateNew_RebuildSpellList()

	-- allow the current selection to remain if it is valid
	if DetectedEvent then
		if self:MatchesFilter(DetectedEvent,false) then
			-- the current selection is valid, so keep it
			return
		end
		--else, the current selection did not match the filter, so select a default
	end

	-- select the first thing in the list that matches the selected filter, if possible
	for key,DetectedEvent in pairs(SpeakinSpellSavedData.NewEventsDetected) do
		if self:MatchesFilter(DetectedEvent,false) then
			-- the current selection is valid, so keep it
			self.OptionsGUIStates.CreateNew.SelectedEventKey = key
			return
		end
	end

	-- our spell list is empty or nothing matches the filter
	self.OptionsGUIStates.CreateNew.SelectedEventKey = nil
end


function SpeakinSpell:CreateNew_OnEventTypeFilterSelect(getset,value)
	if "GET" == getset then
		-- rebuild the list to match the new filter if its out of date
		self:CreateNew_RebuildSpellList()
		return SpeakinSpell.RuntimeData.SelectedEventTypeFilter
	else -- "SET"
		-- if the filter is changing, then rebuild the spell/event list
		self:SetFilter(value,nil)
		-- rebuild the list to match the new filter
		self:CreateNew_RebuildSpellList()
	end
	-- NOTE: the current selection will be repaired to match the new filter in _ValidateSelectedEvent
	--		as a side effect when the GUI framework automatically tries to get the new value of 
	--		the rest of the controls on the page, including the selected event
end


function SpeakinSpell:CreateNew_OnGetSetEventTextFilter(getset,value)
	if "GET" == getset then
		--if nothing is selected, set to all
		self:CreateNew_RebuildSpellList()
		return SpeakinSpell.RuntimeData.SelectedEventTextFilter
	else -- "SET"
		-- if the filter is changing, then rebuild the spell/event list
		self:SetFilter(nil,value)
		-- rebuild the list to match the new filter
		self:CreateNew_RebuildSpellList()
	end
	-- NOTE: the current selection will be repaired to match the new filter in _ValidateSelectedEvent
	--		as a side effect when the GUI framework automatically tries to get the new value of 
	--		the rest of the controls on the page, including the selected event
end



function SpeakinSpell:CreateNew_OnToggleShowAllRanks(getset,value)
	if "GET" == getset then
		-- rebuild the list to match the new filter if its out of date
		self:CurrentMessagesGUI_RebuildSpellList()
		return SpeakinSpellSavedData.ShowAllRanks
	else -- "SET"
		-- if the filter is changing, then rebuild the spell/event list
		self:SetFilterShowAllRanks(value)
		-- rebuild the list to match the new filter
		self:CurrentMessagesGUI_RebuildSpellList()
	end
end


function SpeakinSpell:CreateNew_OnNewSpellNameSelect(getset, value)
	if "GET" == getset then
		--if nothing is selected, try to select the first thing in the list
		self:CreateNew_ValidateSelectedEvent()
		return self.OptionsGUIStates.CreateNew.SelectedEventKey
	else -- "SET"
		self.OptionsGUIStates.CreateNew.SelectedEventKey = value
	end
end


function SpeakinSpell:CreateNew_RebuildSpellList()
	if not self.OptionsGUIStates then
		return
	end
	if not self.OptionsGUIStates.CreateNew.FilterChanged then
		return
	end

	-- reset the list
	self.OptionsGUI.args.CreateNew.args.NewSpellNameSelect.values = {}
	local OldSelectedEvent = SpeakinSpellSavedData.NewEventsDetected[ self.OptionsGUIStates.CreateNew.SelectedEventKey ]
	
	-- search all new events for those that match the filter
	local anykey = ""
	for key,DetectedEvent in pairs(SpeakinSpellSavedData.NewEventsDetected) do
		-- only use events that match the current filter
		if self:MatchesFilter(DetectedEvent,false) then
			-- add to the list of shown values
			self.OptionsGUI.args.CreateNew.args.NewSpellNameSelect.values[key] = DetectedEvent.DisplayName
			-- make sure we have a current spell selected for the options GUI 
			-- we'll use the last one we see because it's easier
			anykey = DetectedEvent.key
		end
	end
	
	if OldSelectedEvent and self:MatchesFilter( OldSelectedEvent,false ) then
		self.OptionsGUIStates.CreateNew.SelectedEventKey = OldSelectedEvent.key
	else
		self.OptionsGUIStates.CreateNew.SelectedEventKey = anykey
	end

	self.OptionsGUIStates.CreateNew.FilterChanged = false
end

function SpeakinSpell:CreateNew_OnClickCreateNew()
	local funcname = "CreateNew_OnClickCreateNew"
	
	-- grab the selected new event
	local de = self:CreateNew_GetSelectedEventObject()
	if not de then
		self:ErrorMsg(funcname, "no event selected")
		return
	end
	
	-- add the spell to the message table and related data structures
	if not SpeakinSpellSavedData.EventTable[de.key] then
		SpeakinSpellSavedData.EventTable[de.key] = {
			-- event information
			DetectedEvent = {
				key			= de.key,
				name		= de.name,
				rank		= de.rank,
				DisplayName = de.DisplayName,
				SpellLink	= de.SpellLink,
				type		= de.type,
				-- caster	= de.caster -- not stored in event table
				-- target	= de.target -- not stored in event table
				-- pet		= de.pet    -- not stored in event table
			},
			-- default options
			Channels = self:CopyTable(SpeakinSpell.DEFAULTS.CHANNEL_RULES),
			Messages = self:CopyTable(DEFAULT_SPEECHES.NEWSPELL),
			DisableAnnouncements = false,
			WhisperTarget = false,
			Frequency = SpeakinSpell.DEFAULTS.FREQUENCY,
			Cooldown  = SpeakinSpell.DEFAULTS.COOLDOWN,
		}
	end
	
	-- destroy the new event detection since this is now a registered event
	-- NO! keep it.
	-- NOTE: We save the whole list now forever and we don't want to have to recast to get it back
	-- SpeakinSpellSavedData.NewEventsDetected[de.key] = nil
	
	-- deselect it from the Create New screen and remove it from that UI
	-- NO! see note above
	-- self.OptionsGUIStates.CreateNew.SelectedEventKey = nil
	-- self.OptionsGUI.args.CreateNew.args.NewSpellNameSelect.values[de.key] = nil

	-- setup the message settings GUI to show the new category of event we've added
	self:SetFilter(de.type,de.name)
	-- rebuild the spell list to include our new event, now that we've selected the appropriate filter
	self:CurrentMessagesGUI_RebuildSpellList()
	-- and select our new spell
	self.OptionsGUIStates.MessageSettings.SelectedEventKey = de.key
	
	-- now we're ready to go to the new event's settings in the options GUI
	InterfaceOptionsFrame_OpenToCategory(L["Message Settings TITLE"])
end


-------------------------------------------------------------------------------
-- OPTIONS GUI FUNCTIONS - RANDOM SUBS PAGE
-------------------------------------------------------------------------------


function SpeakinSpell:RandomSubs_BuildSubsList()
	for key,table in pairs(SpeakinSpellSavedData.RandomSubs) do
		self.OptionsGUI.args.RandomSubs.args.RandomSubSelect.values[key] = "<"..key..">"
	end
end


function SpeakinSpell:RandomSubs_ValidateSelectedRandomSub()
	if self.OptionsGUIStates.RandomSubs.RandomSub and SpeakinSpellSavedData.RandomSubs[ self.OptionsGUIStates.RandomSubs.RandomSub ] then
		return
	end
	
	for key,table in pairs(SpeakinSpellSavedData.RandomSubs) do
		self.OptionsGUIStates.RandomSubs.RandomSub = key
		return
	end
	
	self.OptionsGUIStates.RandomSubs.RandomSub = nil
end


function SpeakinSpell:RandomSubs_GetSelectedSubsTable()
	self:RandomSubs_ValidateSelectedRandomSub()
	local key = self.OptionsGUIStates.RandomSubs.RandomSub
	if key then
		return SpeakinSpellSavedData.RandomSubs[ key ]
	else
		return nil
	end
end


function SpeakinSpell:RandomSubs_OnRandomSubSelect(getset, value)
	if "GET" == getset then
		self:RandomSubs_ValidateSelectedRandomSub()
		for i=1,SpeakinSpell.MAX.TEXTS_PER_RANDOM_SUB,1 do
			self.OptionsGUI.args.RandomSubs.args["RandomSub"..i].name = "<"..self.OptionsGUIStates.RandomSubs.RandomSub.."> "..i
			self.OptionsGUI.args.RandomSubs.args["RandomSub"..i].desc = string.format(L["Anything you type here might be used as a substitution for <%s>"], self.OptionsGUIStates.RandomSubs.RandomSub)
		end
		return self.OptionsGUIStates.RandomSubs.RandomSub
	else
		self.OptionsGUIStates.RandomSubs.RandomSub = value
	end
end


function SpeakinSpell:RandomSubs_OnTextBox(getset, index, value)
	local SubsTable = self:RandomSubs_GetSelectedSubsTable()
	if not SubsTable then
		self:ErrorMsg("RandomSubs_OnTextBox", "no SubsTable selected")
		return ""
	end
	
	if "GET" == getset then
		local text = SubsTable[index]
		if not text then
			text = ""
		end
		return text
	else -- "SET"
		-- store the new value into the message table
		SubsTable[index] = value
		
		-- remove holes from the message table
		-- if the user has just deleted one of the entries, move everything else up to fill in the gap
		if value == "" then
			for i=index+1,SpeakinSpell.MAX.TEXTS_PER_RANDOM_SUB,1 do
				SubsTable[i-1] = SubsTable[i]
			end
			SubsTable[SpeakinSpell.MAX.TEXTS_PER_RANDOM_SUB] = nil
			-- destroy empty strings from the table
			for i=SpeakinSpell.MAX.TEXTS_PER_RANDOM_SUB-1,1,-1 do
				if SubsTable[i] == "" then
					SubsTable[i] = nil
				else
					break --the rest should be valid non-empty strings
				end
			end
		end
	end
end


function SpeakinSpell:RandomSubs_ShowTextBox(i)
	local NumTextBoxesShown = 0
	local SubsTable = self:RandomSubs_GetSelectedSubsTable()
	if SubsTable then
		NumTextBoxesShown = 1 + #(SubsTable)
	end

	return (i <= NumTextBoxesShown)
end


-------------------------------------------------------------------------------
-- OPTIONS GUI FUNCTIONS - HELP PAGES
-------------------------------------------------------------------------------


function SpeakinSpell:HelpPages_Build()
	-- build help pages
	--for title,contents in pairs(HELPFILE.PAGES) do --has an unpredictable order
	for i=1,#(HELPFILE.PAGEORDER),1 do
		local FrameObjectName = "HelpTopic"..tostring(i)
		local FrameDisplayName = HELPFILE.PAGEORDER[i]
		local NextPageDisplayName = ""
		if i < #(HELPFILE.PAGEORDER) then
			NextPageDisplayName = HELPFILE.PAGEORDER[i+1]
		else --wrap to first page
			NextPageDisplayName = HELPFILE.PAGEORDER[1]
		end
		-- setup OptionGUI object
		self.OptionsGUI.args[FrameObjectName] = {
			order = 3,
			type = "group",
			name = "unused_HelpFrame_name",
			desc = "unused_HelpFrame_desc",
			args = {
				CaptionNormalMode = {
					order = 1,
					type = "description",
					name = HELPFILE.PAGES[ HELPFILE.PAGEORDER[i] ], --help CONTENTS
				},
				NextPage = {
					order = 2,
					type = "execute",
					width = "full",
					name = NextPageDisplayName,
					desc = L["Go to the next help topic"],
					func = function() InterfaceOptionsFrame_OpenToCategory(NextPageDisplayName) end
				}
			}, --end args
		} --end title
		
		-- add a button to the TOC page, linking to the child page we just created above
		self.OptionsGUI.args.HelpRoot.args[FrameObjectName] = {
			order = i+1,
			type = "execute",
			width = "full",
			name = FrameDisplayName,
			desc = FrameDisplayName, --L["Go to the first help topic"], --TODOLATER: tooltips different from labels, and cleanup L[] entry
			func = function() InterfaceOptionsFrame_OpenToCategory(FrameDisplayName) end
		}

	end
end


function SpeakinSpell:HelpPages_AddToBlizOptions()
	--for title,contents in pairs(HELPFILE.PAGES) do --has an unpredictable order
	for i=1,#(HELPFILE.PAGEORDER),1 do
		local FrameObjectName = "HelpTopic"..tostring(i)
		local FrameDisplayName = HELPFILE.PAGEORDER[i]
		--TODOFUTURE: push down to tier 3 -- tier 3 seems buggy, trying tier 2 instead
		--self:AddToBlizOptions(FrameObjectName, FrameDisplayName, L["SpeakinSpell Help"])
		self:AddToBlizOptions(FrameObjectName, FrameDisplayName, L["SpeakinSpell"])
	end
end


-------------------------------------------------------------------------------
-- OPTIONS GUI FUNCTIONS - SHARED UTILITIES
-------------------------------------------------------------------------------


function SpeakinSpell:AddToBlizOptions(FrameObjectName, FrameDisplayName, ParentDisplayName)
	-- AddToBlizOptions(appName="SpeakinSpell", display name, parent, ObjectName)
	-- NOTE: the FrameObjectName here represents self.OptionsGUI.args[FrameObjectName]
	self.optionsFrames[FrameObjectName] = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("SpeakinSpell", FrameDisplayName, ParentDisplayName, FrameObjectName)
	
	-- for some reason the framework has a problem if we define the "reset defaults" button handler in the data before calling AddToBlizOptions
	-- so we have to add it afterwards
	self.optionsFrames[FrameObjectName].default = function() self:ResetToDefaults(FrameObjectName) end
end


function SpeakinSpell:InitOptionsGUIState() -- sets up event lists and current selections
	--declare variables so they exist
	self.OptionsGUIStates = {
		MessageSettings = {
			SelectedEventKey = nil,
			FilterChanged = true,
		},
		CreateNew = {
			SelectedEventKey = nil,
			FilterChanged = true,
		},
		RandomSubs = {
			SelectedSub = nil,
		},
	}

	--reset the search filter
	self:SetFilter("ALL","")

	-- the options GUI may still have a reference to the old SavedData table.
	-- refresh it since we've rebuilt the table
	-- but if we're initializing then the OptionsGUI may not have been created yet
	if self.OptionsGUI then
		self:CurrentMessagesGUI_RebuildSpellList()
		self:CreateNew_RebuildSpellList()
		self:RandomSubs_BuildSubsList()
	end
end


function SpeakinSpell:SetupOptionsGUI_CreateMessageInputs()
	-- add SpeakinSpell.MAX.MESSAGES_PER_SPELL random speech input boxes
	for i=1,SpeakinSpell.MAX.MESSAGES_PER_SPELL,1 do
		self.OptionsGUI.args.CurrentMessagesGUI.args["SpellMessage"..i] = {
			order = i + 100, -- should be the order of the last item listed above the message input boxes (OncePerTargetToggle), plus i
			width = "full",
			type = "input",
			name = string.format(L["Random Speech i"], i), -- "Random Speech %d"
			desc = L["Random Speech i TOOLTIP"],
			get = function() return  self:CurrentMessagesGUI_OnMessage("GET",i,nil) end,
			set = function(_, value) self:CurrentMessagesGUI_OnMessage("SET",i, value) end,
			hidden = function() return not self:CurrentMessagesGUI_ShowSpeechBox(i) end,
		}
	end
end


function SpeakinSpell:SetupOptionsGUI_CreateRandomSubsInputs()
	for i=1,SpeakinSpell.MAX.TEXTS_PER_RANDOM_SUB,1 do
		self.OptionsGUI.args.RandomSubs.args["RandomSub"..i] = {
			order = i + 100, -- should be the order of the last item listed above the message input boxes (OncePerTargetToggle), plus i
			width = "full",
			type = "input",
			name = "", --the name and desc tooltip are generated dynamically to use the selected <randomtext> variable name
			desc = "",
			get = function() return  self:RandomSubs_OnTextBox("GET",i,nil) end,
			set = function(_, value) self:RandomSubs_OnTextBox("SET",i, value) end,
			hidden = function() return not self:RandomSubs_ShowTextBox(i) end,
		}
	end
end


function SpeakinSpell:RefreshFrame(FrameName)
	-- the only way I found to do this is to hide and then show the frame
	-- this occurs fast enough that I do not see a flicker when this happens
	-- UPDATE: saw a forum post when I was at work with the proper API call, but I already forget what it was, and this still works
	if self.optionsFrames and self.optionsFrames[FrameName] and self.optionsFrames[FrameName]:IsVisible() then
		self.optionsFrames[FrameName]:Hide()
		self.optionsFrames[FrameName]:Show()
	end
end


-------------------------------------------------------------------------------
-- OPTIONS GUI FUNCTIONS - ENTRY POINT FOR ALL
-------------------------------------------------------------------------------


function SpeakinSpell:SetupOptionsGUI()
	-- Define Options table for the GUI
	self.OptionsGUI = {
		type = "group",
		name = L["SpeakinSpell"],
		args = {
	---------------------------------------------------------------------------------------
			General = {
				order = 1,
				type = "group",
				name = L["General Settings"],
				desc = L["General Settings TOOLTIP"],
				args = {
					Caption = {
						order = 1,
						type = "description",
						name = L["General Settings"],
					},
					ShowVersion = {
						order = 2,
						type = "description",
						name = L["Version "]..SpeakinSpell.CURRENT_VERSION,
					},
					ShowURL = {
						order = 3,
						type = "description",
						name = SpeakinSpell.URL,
					},
					EnableAllMessagesToggle = {
						order = 4,
						type = "toggle",
						width = "full",
						name = L["Enable All Messages"],
						desc = L["Enable All Messages TOOLTIP"],
						get = function() return  self:GeneralOptions_OnEnableAllMessagesToggle("GET",nil) end,
						set = function(_, value) self:GeneralOptions_OnEnableAllMessagesToggle("SET",value) end,
					},
					ShowSpellEventsToggle = {
						order = 5,
						type = "toggle",
						width = "full",
						name = L["Report detected spell events"],
						desc = L["Report detected spell events TOOLTIP"],
						get = function() return  self:GeneralOptions_OnShowSpellEventsToggle("GET",nil) end,
						set = function(_, value) self:GeneralOptions_OnShowSpellEventsToggle("SET",value) end,
					},
					ShowDebugMessagesToggle = {
						order = 6,
						type = "toggle",
						width = "full",
						name = L["Show Debugging Messages (verbose)"],
						desc = L["Show Debugging Messages TOOLTIP"],
						get = function() return  self:GeneralOptions_OnShowDebugMessagesToggle("GET",nil) end,
						set = function(_, value) self:GeneralOptions_OnShowDebugMessagesToggle("SET",value) end,
					},
					ImportDefaultsButton = {
						order = 7,
						type = "execute",
						width = "full",
						name = L["Import New Default Sample Speeches"],
						desc = L["Import New Default Sample Speeches TOOLTIP"],
						func = function() self:GeneralOptions_OnClickImport() end,
					},
					ScrollGap = {
						order = 8,
						type = "description",
						name = "\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n",
						func = function() self:GeneralOptions_OnClickImport() end,
					},
					BigRedButton = {
						order = 9,
						type = "execute",
						width = "full",
						name = L["DO NOT PRESS THIS BUTTON"],
						desc = L["Under no circumstances should you ever, EVER push this button.\n\nI'm warning you.\n\nDon't do it.\n\nSeriously, you're already hovering over it for too long, and that's dangerous.\n\nMove your mouse before it's too late!"],
						func = function() self:GeneralOptions_OnClickBigRedButton() end,
					},
				}, --end args
			}, --end General

	---------------------------------------------------------------------------------------
			CurrentMessagesGUI = {
				order = 1,
				type = "group",
				name = L["General Settings"],
				desc = L["General Settings TOOLTIP"],
				args = {
					CaptionNormalMode = {
						order = 1,
						type = "description",
						name = L["Message Settings"],
					},
					EventTextFilterSelect = {
						order = 2,
						type = "input",
						width = "full",
						name = L["Search"],
						desc = L["Type some of the name of what you're looking for, to narrow down the list below"],
						get = function() return  self:CurrentMessagesGUI_OnGetSetEventTextFilter("GET",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnGetSetEventTextFilter("SET",value) end,
					},
					EventTypeFilterSelect = {
						order = 3,
						type = "select",
						width = "full",
						name = L["Select a Category of Events"],
						desc = L["Select a Category of Events TOOLTIP"],
						values = L.EVENTTYPES.AS_FILTERS,
						get = function() return  self:CurrentMessagesGUI_OnEventTypeFilterSelect("GET",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnEventTypeFilterSelect("SET",value) end,
					},
					EventSelect = {
						order = 4,
						type = "select",
						width = "full",
						name = L["Select a Spell"],
						desc = L["Select a Spell TOOLTIP"],
						values = {},
						get = function() return  self:CurrentMessagesGUI_OnEventSelect("GET",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnEventSelect("SET",value) end,
						hidden = function() return not self:CurrentMessagesGUI_ShowSelectedEventControls() end,
					},
					CreateSpellButton = {
						order = 5,
						type = "execute",
						name = L["Create New..."],
						desc = L["Create New... TOOLTIP"],
						func = function() self:CurrentMessagesGUI_OnClickGotoCreate() end,
					},
					DeleteSpellButton = {
						order = 5,
						type = "execute",
						name = L["Delete Spell"],
						desc = L["Delete Spell TOOLTIP"],
						func = function() self:CurrentMessagesGUI_OnClickDelete() end,
						hidden = function() return not self:CurrentMessagesGUI_ShowSelectedEventControls() end,
					},
					Disable = {
						order = 6,
						type = "toggle",
						width = "full",
						name = L["Disable announcements for this spell"],
						desc = L["Disable announcements for this spell TOOLTIP"],
						get = function() return  self:CurrentMessagesGUI_OnSpellDisable("GET",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnSpellDisable("SET",value) end,
						hidden = function() return not self:CurrentMessagesGUI_ShowSelectedEventControls() end,
					},
					ChannelSelectCaption = {
						order = 7,
						type = "description",
						width= "full",
						name = L["Select the channels to use when announcing this spell when"],
						hidden = function() return not self:CurrentMessagesGUI_ShowSelectedEventControls() end,
					},
					ChannelSelectSolo = {
						order = 8,
						type = "select",
						name = L["By yourself"],
						desc = L["By yourself TOOLTIP"],
						hidden = function() return not self:CurrentMessagesGUI_ShowSelectedEventControls() end,
						values = { -- color-coded to match default chat channel colors
							[L["SPEAKINSPELL CHANNEL"]]	= "|cff33ff99"..L["SPEAKINSPELL CHANNEL"],
							[L["GUILD"]]		= "|c4444ff44"..L["GUILD"],
							--[L["BATTLEGROUND"]]	= "|cffff6b00"..L["BATTLEGROUND"],
							--[L["RAID"]]			= "|cffff6b00"..L["RAID"],
							--[L["PARTY"]]		= "|cffaaeeff"..L["PARTY"],
							[L["SAY"]]			=               L["SAY"],
							[L["EMOTE"]]		= "|cffff7b34"..L["EMOTE"],
							[L["YELL"]]			= "|cffff2f32"..L["YELL"],
							--[L["RAID_WARNING"]]	= "|cffffdbad"..L["RAID_WARNING"],
							[L["Silent"]]		=               L["Silent"], --tried light gray to indicate disabled, but it made the control appear to be disabled, so use the default color (white)
						},
						get = function() return  self:CurrentMessagesGUI_OnChannelSelect("GET","Solo",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnChannelSelect("SET","Solo",value) end,
					},
					ChannelSelectParty = {
						order = 9,
						type = "select",
						name = L["In a Party"],
						desc = L["In a Party TOOLTIP"],
						hidden = function() return not self:CurrentMessagesGUI_ShowSelectedEventControls() end,
						values = { -- color-coded to match default chat channel colors
							--[L["BATTLEGROUND"]]	= "|cffff6b00"..L["BATTLEGROUND"],
							--[L["RAID"]]			= "|cffff6b00"..L["RAID"],
							[L["SPEAKINSPELL CHANNEL"]]	= "|cff33ff99"..L["SPEAKINSPELL CHANNEL"],
							[L["GUILD"]]		= "|c4444ff44"..L["GUILD"],
							[L["PARTY"]]		= "|cffaaeeff"..L["PARTY"],
							[L["SAY"]]			=               L["SAY"],
							[L["EMOTE"]]		= "|cffff7b34"..L["EMOTE"],
							[L["YELL"]]			= "|cffff2f32"..L["YELL"],
							[L["RAID_WARNING"]]	= "|cffffdbad"..L["RAID_WARNING"],
							[L["Silent"]]		=               L["Silent"], --tried light gray to indicate disabled, but it made the control appear to be disabled, so use the default color (white)
						},
						get = function() return  self:CurrentMessagesGUI_OnChannelSelect("GET","Party",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnChannelSelect("SET","Party",value) end,
					},
					ChannelSelectRaid = {
						order = 10,
						type = "select",
						name = L["In a Raid"],
						desc = L["In a Raid TOOLTIP"],
						hidden = function() return not self:CurrentMessagesGUI_ShowSelectedEventControls() end,
						values = { -- color-coded to match default chat channel colors
							--[L["BATTLEGROUND"]]	= "|cffff6b00"..L["BATTLEGROUND"],
							[L["SPEAKINSPELL CHANNEL"]]	= "|cff33ff99"..L["SPEAKINSPELL CHANNEL"],
							[L["GUILD"]]		= "|c4444ff44"..L["GUILD"],
							[L["RAID"]]			= "|cffff6b00"..L["RAID"],
							[L["PARTY"]]		= "|cffaaeeff"..L["PARTY"],
							[L["SAY"]]			=               L["SAY"],
							[L["EMOTE"]]		= "|cffff7b34"..L["EMOTE"],
							[L["YELL"]]			= "|cffff2f32"..L["YELL"],
							[L["RAID_WARNING"]]	= "|cffffdbad"..L["RAID_WARNING"],
							[L["Silent"]]		=               L["Silent"], --tried light gray to indicate disabled, but it made the control appear to be disabled, so use the default color (white)
						},
						get = function() return  self:CurrentMessagesGUI_OnChannelSelect("GET","Raid",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnChannelSelect("SET","Raid",value) end,
					},
					ChannelSelectPartyLeader = {
						order = 11,
						type = "select",
						name = L["Party Leader"],
						desc = L["Use this channel if you are the leader of a 5-man party"],
						hidden = function() return not self:CurrentMessagesGUI_ShowSelectedEventControls() end,
						values = { -- color-coded to match default chat channel colors
							--[L["BATTLEGROUND"]]	= "|cffff6b00"..L["BATTLEGROUND"],
							[L["SPEAKINSPELL CHANNEL"]]	= "|cff33ff99"..L["SPEAKINSPELL CHANNEL"],
							[L["GUILD"]]		= "|c4444ff44"..L["GUILD"],
							[L["RAID"]]			= "|cffff6b00"..L["RAID"],
							[L["PARTY"]]		= "|cffaaeeff"..L["PARTY"],
							[L["SAY"]]			=               L["SAY"],
							[L["EMOTE"]]		= "|cffff7b34"..L["EMOTE"],
							[L["YELL"]]			= "|cffff2f32"..L["YELL"],
							[L["RAID_WARNING"]]	= "|cffffdbad"..L["RAID_WARNING"],
							[L["Silent"]]		=               L["Silent"], --tried light gray to indicate disabled, but it made the control appear to be disabled, so use the default color (white)
						},
						get = function() return  self:CurrentMessagesGUI_OnChannelSelect("GET","PartyLeader",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnChannelSelect("SET","PartyLeader",value) end,
					},
					ChannelSelectRaidLeader = {
						order = 12,
						type = "select",
						name = L["Raid Leader"],
						desc = L["Use this channel if you are the leader of a raid group"],
						hidden = function() return not self:CurrentMessagesGUI_ShowSelectedEventControls() end,
						values = { -- color-coded to match default chat channel colors
							--[L["BATTLEGROUND"]]	= "|cffff6b00"..L["BATTLEGROUND"],
							[L["SPEAKINSPELL CHANNEL"]]	= "|cff33ff99"..L["SPEAKINSPELL CHANNEL"],
							[L["GUILD"]]		= "|c4444ff44"..L["GUILD"],
							[L["RAID"]]			= "|cffff6b00"..L["RAID"],
							[L["PARTY"]]		= "|cffaaeeff"..L["PARTY"],
							[L["SAY"]]			=               L["SAY"],
							[L["EMOTE"]]		= "|cffff7b34"..L["EMOTE"],
							[L["YELL"]]			= "|cffff2f32"..L["YELL"],
							[L["RAID_WARNING"]]	= "|cffffdbad"..L["RAID_WARNING"],
							[L["Silent"]]		=               L["Silent"], --tried light gray to indicate disabled, but it made the control appear to be disabled, so use the default color (white)
						},
						get = function() return  self:CurrentMessagesGUI_OnChannelSelect("GET","RaidLeader",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnChannelSelect("SET","RaidLeader",value) end,
					},
					ChannelSelectRaidOfficer = {
						order = 13,
						type = "select",
						name = L["Raid Officer"],
						desc = L["Use this channel if you are promoted to assist in a raid group"],
						hidden = function() return not self:CurrentMessagesGUI_ShowSelectedEventControls() end,
						values = { -- color-coded to match default chat channel colors
							--[L["BATTLEGROUND"]]	= "|cffff6b00"..L["BATTLEGROUND"],
							[L["SPEAKINSPELL CHANNEL"]]	= "|cff33ff99"..L["SPEAKINSPELL CHANNEL"],
							[L["GUILD"]]		= "|c4444ff44"..L["GUILD"],
							[L["RAID"]]			= "|cffff6b00"..L["RAID"],
							[L["PARTY"]]		= "|cffaaeeff"..L["PARTY"],
							[L["SAY"]]			=               L["SAY"],
							[L["EMOTE"]]		= "|cffff7b34"..L["EMOTE"],
							[L["YELL"]]			= "|cffff2f32"..L["YELL"],
							[L["RAID_WARNING"]]	= "|cffffdbad"..L["RAID_WARNING"],
							[L["Silent"]]		=               L["Silent"], --tried light gray to indicate disabled, but it made the control appear to be disabled, so use the default color (white)
						},
						get = function() return  self:CurrentMessagesGUI_OnChannelSelect("GET","RaidOfficer",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnChannelSelect("SET","RaidOfficer",value) end,
					},
					ChannelSelectBG = {
						order = 14,
						type = "select",
						name = L["In a Battleground"],
						desc = L["In a Battleground TOOLTIP"],
						hidden = function() return not self:CurrentMessagesGUI_ShowSelectedEventControls() end,
						values = { -- color-coded to match default chat channel colors
							[L["SPEAKINSPELL CHANNEL"]]	= "|cff33ff99"..L["SPEAKINSPELL CHANNEL"],
							[L["GUILD"]]		= "|c4444ff44"..L["GUILD"],
							[L["BATTLEGROUND"]]	= "|cffff6b00"..L["BATTLEGROUND"],
							--[L["RAID"]]			= "|cffff6b00"..L["RAID"],
							--[L["PARTY"]]		= "|cffaaeeff"..L["PARTY"],
							[L["SAY"]]			=               L["SAY"],
							[L["EMOTE"]]		= "|cffff7b34"..L["EMOTE"],
							[L["YELL"]]			= "|cffff2f32"..L["YELL"],
							--[L["RAID_WARNING"]]	= "|cffffdbad"..L["RAID_WARNING"],
							[L["Silent"]]		=               L["Silent"], --tried light gray to indicate disabled, but it made the control appear to be disabled, so use the default color (white)
						},
						get = function() return  self:CurrentMessagesGUI_OnChannelSelect("GET","BG",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnChannelSelect("SET","BG",value) end,
					},
					ChannelSelectArena = {
						order = 15,
						type = "select",
						name = L["In Arena"],
						desc = L["In Arena TOOLTIP"],
						hidden = function() return not self:CurrentMessagesGUI_ShowSelectedEventControls() end,
						values = { -- color-coded to match default chat channel colors
							[L["SPEAKINSPELL CHANNEL"]]	= "|cff33ff99"..L["SPEAKINSPELL CHANNEL"],
							[L["GUILD"]]		= "|c4444ff44"..L["GUILD"],
							--[L["BATTLEGROUND"]]	= "|cffff6b00"..L["BATTLEGROUND"],
							[L["RAID"]]			= "|cffff6b00"..L["RAID"],
							[L["PARTY"]]		= "|cffaaeeff"..L["PARTY"],
							[L["SAY"]]			=               L["SAY"],
							[L["EMOTE"]]		= "|cffff7b34"..L["EMOTE"],
							[L["YELL"]]			= "|cffff2f32"..L["YELL"],
							[L["RAID_WARNING"]]	= "|cffffdbad"..L["RAID_WARNING"],
							[L["Silent"]]		=               L["Silent"], --tried light gray to indicate disabled, but it made the control appear to be disabled, so use the default color (white)
						},
						get = function() return  self:CurrentMessagesGUI_OnChannelSelect("GET","Arena",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnChannelSelect("SET","Arena",value) end,
					},
					ChannelSelectWG = {
						order = 16,
						type = "select",
						name = L["In Wintergrasp"],
						desc = L["In Wintergrasp TOOLTIP"],
						hidden = function() return not self:CurrentMessagesGUI_ShowSelectedEventControls() end,
						values = { -- color-coded to match default chat channel colors
							[L["SPEAKINSPELL CHANNEL"]]	= "|cff33ff99"..L["SPEAKINSPELL CHANNEL"],
							[L["GUILD"]]		= "|c4444ff44"..L["GUILD"],
							--[L["BATTLEGROUND"]]	= "|cffff6b00"..L["BATTLEGROUND"],
							[L["RAID"]]			= "|cffff6b00"..L["RAID"],
							[L["PARTY"]]		= "|cffaaeeff"..L["PARTY"],
							[L["SAY"]]			=               L["SAY"],
							[L["EMOTE"]]		= "|cffff7b34"..L["EMOTE"],
							[L["YELL"]]			= "|cffff2f32"..L["YELL"],
							--[L["RAID_WARNING"]]	= "|cffffdbad"..L["RAID_WARNING"],
							[L["Silent"]]		=               L["Silent"], --tried light gray to indicate disabled, but it made the control appear to be disabled, so use the default color (white)
						},
						get = function() return  self:CurrentMessagesGUI_OnChannelSelect("GET","WG",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnChannelSelect("SET","WG",value) end,
					},
					WhisperTargetToggle = {
						order = 17,
						type = "toggle",
						width = "full",
						name = L["Whisper the message to your target"],
						desc = L["Whisper the message to your target TOOLTIP"],
						get = function() return  self:CurrentMessagesGUI_OnWhisperTarget("GET",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnWhisperTarget("SET",value) end,
						hidden = function() return not self:CurrentMessagesGUI_ShowSelectedEventControls() end,
					},
					FrequencySelect = {
						order = 18,
						type = "range",
						width = "full",
						name = L["How often do you want to announce this spell?"],
						desc = L["How often TOOLTIP"],
						min = 0.01, max = 1.00, step=0.01,
						isPercent = true,
						get = function() return  self:CurrentMessagesGUI_OnFrequencySelect("GET",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnFrequencySelect("SET",value) end,
						hidden = function() return not self:CurrentMessagesGUI_ShowSelectedEventControls() end,
					},
					CooldownSlider = {
						order = 19,
						type = "range",
						width = "full",
						name = L["Cooldown on announcements for this spell"] ,
						desc = L["Cooldown TOOLTIP"],
						min =0, max = 600, step=1, -- allowing for 0 seconds up to 10 minutes
						get = function() return  self:CurrentMessagesGUI_OnCooldownSlider("GET",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnCooldownSlider("SET",value) end,
						hidden = function() return not self:CurrentMessagesGUI_ShowSelectedEventControls() end,
					},
					OncePerCombatToggle = {
						order = 20,
						type = "toggle",
						width = "double",
						name = L["Limit once per combat"],
						desc = L["Limit once per combat TOOLTIP"],
						get = function() return  self:CurrentMessagesGUI_OnOncePerCombatToggle("GET",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnOncePerCombatToggle("SET",value) end,
						hidden = function() return not self:CurrentMessagesGUI_ShowSelectedEventControls() end,
					},
					OncePerTargetToggle = {
						order = 21,
						type = "toggle",
						width = "double",
						name = L["Limit once per target"],
						desc = L["Limit once per target TOOLTIP"],
						get = function() return  self:CurrentMessagesGUI_OnOncePerTargetToggle("GET",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnOncePerTargetToggle("SET",value) end,
						hidden = function() return not self:CurrentMessagesGUI_ShowSelectedEventControls() end,
					},
				}, --end args
			}, --end CurrentMessagesGUI

	---------------------------------------------------------------------------------------
			RandomSubs = {
				order = 1,
				type = "group",
				name = L["Random Substitutions"],
				desc = L["Random Substitutions like <randomtaunt> and <randomfaction>"],
				args = {
					CaptionNormalMode = {
						order = 1,
						type = "description",
						name = L["Random Substitutions"],
					},
					Instructions = {
						order = 2,
						type = "description",
						name = "\n\n" .. L["Use this interface to change the values which may be used for the randomized substitutions, <randomtaunt> and <randomfaction>"] .. "\n\n",
					},
					RandomSubSelect = {
						order = 3,
						type = "select",
						width = "full",
						name = L["Select a Randomized Substitution"],
						desc = L["Select a Randomized Substitution\n\nThese substitutions may be used in speeches when announcing events, and a random result from the list below will be selected."],
						values = {},
						get = function() return  self:RandomSubs_OnRandomSubSelect("GET",nil) end,
						set = function(_, value) self:RandomSubs_OnRandomSubSelect("SET",value) end,
					},
				}, --end args
			}, --end Trade
			
	---------------------------------------------------------------------------------------
			Trade = {
				order = 1,
				type = "group",
				name = "unused_Trade_name",
				desc = "unused_Trade_desc",
				args = {
					CaptionNormalMode = {
						order = 1,
						type = "description",
						name = "Coming Soon... randomized trade spam for all of your professional and class services, portals, LFG, LFW",
					},
				}, --end args
			}, --end Trade
			
	---------------------------------------------------------------------------------------
			CreateNew = {
				order = 1,
				type = "group",
				name = "unused_CreateNew_name",
				desc = "unused_CreateNew_desc",
				args = {
					NewSpellModeInstruct = {
						order = 1,
						type = "description",
						name = L["NEW SPELL INSTRUCTIONS"],
					},
					EventTextFilterSelect = {
						order = 2,
						type = "input",
						width = "full",
						name = L["Search"],
						desc = L["Type some of the name of what you're looking for, to narrow down the list below"],
						get = function() return  self:CreateNew_OnGetSetEventTextFilter("GET",nil) end,
						set = function(_, value) self:CreateNew_OnGetSetEventTextFilter("SET",value) end,
					},
					EventTypeFilterSelect = {
						order = 3,
						type = "select",
						width = "full",
						name = L["Select a Category of Events"],
						desc = L["Select a Category of Events TOOLTIP"],
						values = L.EVENTTYPES.AS_FILTERS,
						get = function() return  self:CreateNew_OnEventTypeFilterSelect("GET",nil) end,
						set = function(_, value) self:CreateNew_OnEventTypeFilterSelect("SET",value) end,
					},
					ShowAllRanks = {
						order = 3,
						type = "toggle",
						width = "full",
						name = L["Show All Ranks"],
						desc = L["Enable to show different ranks of spells and abilities, including Polymorph critters.\n\nIf disabled, any rank of the spell will count."],
						hidden = function() return not self:IsSelectedEventTypeFilterRankable() end,
						get = function() return  self:CreateNew_OnToggleShowAllRanks("GET",nil) end,
						set = function(_, value) self:CreateNew_OnToggleShowAllRanks("SET",value) end,
					},
					NewSpellNameSelect = {
						order = 4,
						type = "select",
						width = "full",
						name = L["Select a Spell"],
						desc = L["Select a Spell TOOLTIP - Create New"],
						values = {},
						get = function() return  self:CreateNew_OnNewSpellNameSelect("GET",nil) end,
						set = function(_, value) self:CreateNew_OnNewSpellNameSelect("SET",value) end,
					},
					CreateNewSpellButton = {
						order = 5,
						type = "execute",
						width = "full",
						name = L["Create Selected Event"],
						desc = L["Create Selected Event TOOLTIP"],
						func = function() self:CreateNew_OnClickCreateNew() end,
					},
				}, --end args
			}, --end CreateNew

	---------------------------------------------------------------------------------------
			HelpRoot = {
				order = 2,
				type = "group",
				name = "unused_HelpRoot_name",
				desc = "unused_HelpRoot_desc",
				args = {
					CaptionNormalMode = {
						order = 1,
						type = "description",
						name = HELPFILE.ROOT,
					},
				}, --end args
			}, --end HelpRoot
			
		}, --end args
	} -- end OptionsGUI
	---------------------------------------------------------------------------------------
	
	-- set up spell/event lists and current selections
	self:InitOptionsGUIState()
	
	-- add SpeakinSpell.MAX.MESSAGES_PER_SPELL random speech input boxes
	self:SetupOptionsGUI_CreateMessageInputs()
	self:SetupOptionsGUI_CreateRandomSubsInputs()
	
	-- build help pages
	self:HelpPages_Build()
	
	-- add the options to the in-game interface options panel at Escape > Interface > Addons > SpeakinSpell
	-- The ordering here matters, it determines the order in the Blizzard Interface Options
	-- (or it would matter if we had sub-categories, which we currently don't)
	-- InterfaceOptions_AddCategory(self.OptionsGUI) -- done via Ace below

	-- :RegisterOptionsTable(appName, options, slashcmd, persist)
	-- AceConfig:RegisterOptionsTable("MyAddon", myOptions, {"/myslash", "/my"})
	LibStub("AceConfig-3.0"):RegisterOptionsTable("SpeakinSpell", self.OptionsGUI, nil)
	
	--SpeakinSpell:AddToBlizOptions(FrameObjectName, FrameDisplayName, ParentDisplayName)
	--frame objects are stored to self.optionsFrames[FrameObjectName]
	self.optionsFrames = {}
	self:AddToBlizOptions("General",		L["SpeakinSpell"],		nil)
	self:AddToBlizOptions("CurrentMessagesGUI",	L["Message Settings TITLE"],		L["SpeakinSpell"])
	self:AddToBlizOptions("CreateNew",		L["Create New..."],			L["SpeakinSpell"])
	self:AddToBlizOptions("RandomSubs",		L["Random Substutitions"],	L["SpeakinSpell"])
	self:AddToBlizOptions("Trade",			L["Trade"],					L["SpeakinSpell"])
	self:AddToBlizOptions("HelpRoot",		L["SpeakinSpell Help"],		L["SpeakinSpell"])
	
	-- add help pages
	self:HelpPages_AddToBlizOptions()

end


